//redFrik - released under gnu gpl license
//version 100505

//--changes 100505;
//added verbose classvar and some postlines
//bugfix for empty objects
//--changes 090922:
//bugfix for empty comments
//--changes 090916:
//fix for embedding bpatchers
//added more special characters
//--changes 090713:
//fixed helpfile random comments example
//added more special chars from p.delges - thanks
//--changes 090526:
//added some special chars
//--changes 090525:
//bugfix for coll with filename argument - thanks p.delges
//prevented error while reading patches without newline as last character
//--changes 080712:
//added *new and many helper methods to make patches from scratch
//added setter for name
//--changes 080710:
//fixed serious bug in object. they didn't hide on lock
//added flag for removing bpatcher extensions
//fixed typo bug in ubutton-hilite. magic.set should be magic.setBit
//added setter for dict. so now one can manipulate the maxpat
//wrote asMax5 and saveAsMax5 methods
//major rewrite to prRead. simplified to allow for asMax5
//major rewrite of escape char methods. simplified a lot
//recursive helper method for all lines and all boxes
//--changes 080705:
//removed getter and setter for max4colorsLed. was pointless to have access to it.
//added name instance variable
//replaced .minIndex with manual search. hopefully now works on older versions of sc
//fixed serious bug in object and led colour lookup
//added introspection examples to the help file
//panel gradient mode colours now works correctly

//--todo:
//read compressed - but which scheme are they using?
//read mxf max4 and max5, read clct
//embedded fpic (vpicture)
//workaround for 8192 characters limitation in interpret
//remove f0index from dict when asMax5
//all special characters will not make it thru. how to fix? - needs >1 byte unicode
//test makePatcher with some inlets/outlets

MaxPat {
	classvar >verbose= false;
	var <>dict;									//the max5 patch as a sc dictionary
	var <>name;									//name of patch. usually current file name
	var <>scaleBoxWidth= 1;							//setting to scale all max4 boxes width
	var <>scalePatWidth= 1;							//setting to scale all max4 patches width
	var <>scalePatHeight= 1;						//setting to scale all max4 patches height
	var <>stripExtension= true;						//remove .maxpat extension for bpatchers
	var z= ";\n";
	var lastFontName, lastFontSize, lineCount= false;
	var <>max4colors= #[							//default max colours 0-15
		[170, 170, 170],
		[204, 156, 97],
		[255, 189, 156],
		[204, 140, 140],
		[255, 156, 156],
		[255, 227, 23],
		[40, 204, 140],
		[74, 156, 97],
		[97, 156, 156],
		[156, 179, 255],
		[194, 181, 207],
		[153, 153, 255],
		[102, 102, 204],
		[153, 102, 153],
		[255, 92, 174],
		[255, 138, 215]
	];
	var max4colorsLed= #[							//colours for led 0-4
		[255, 0, 0],
		[0, 255, 0],
		[0, 255, 255],
		[255, 255, 0],
		[255, 255, 255]
	];
	*new {|name, rect, fontName= "Arial", fontSize= 12|
		^super.new.initMaxPat(name, rect, fontName, fontSize);
	}
	initMaxPat {|argName, rect, fontName, fontSize|
		name= (argName ? "untitled");
		rect= (rect ? Rect(25, 69, 640, 506));
		dict= (
			'patcher' : (
				'fileversion' : 1,
				'rect' : rect.asCollection,
				'bglocked' : 0,
				'defrect' : rect.asCollection,
				'openrect' : [0, 0, 0, 0],
				'openinpresentation' : 0,
				'default_fontsize' : fontSize,
				'default_fontface' : 0,
				'default_fontname' : fontName,
				'gridonopen' : 0,
				'gridsize' : [15, 15],
				'gridsnaponopen' : 0,
				'toolbarvisible' : 1,
				'boxanimatetime' : 200,
				'imprint' : 0,
				'metadata' : [],
				'boxes' : [],
				'lines' : []
			)
		);
	}
	*read {|path|
		^super.new.read(path)
	}
	read {|path|
		path= path.standardizePath;
		if(File.exists(path), {
			this.prRead(path);
		}, {(this.class.name++": file"+path+"not found").error});
	}
	asMax5 {										//returns a json max5 patch as a string
		^""++123.asAscii++Char.nl++this.prAsMax5("", dict, 0);
	}
	saveAsMax5 {|path, overwrite= false|
		var file, str;
		path= path.standardizePath;
		if(File.exists(path), {
			if(overwrite==false, {
				(this.class.name++": file"+path+"already exists").error;
				^this;
			});
		});
		str= this.asMax5;
		file= File(path, "w");
		file.write(str);
		file.close;
	}
	asMax4 {										//returns a max4 patch as a string
		^"max v2"++z++this.prAsMax4(dict['patcher']);
	}
	saveAsMax4 {|path, overwrite= false|
		var file, str;
		path= path.standardizePath;
		if(File.exists(path), {
			if(overwrite==false, {
				(this.class.name++": file"+path+"already exists").error;
				^this;
			});
		});
		str= this.asMax4;
		file= File(path, "w");
		file.write(str);
		file.close;
	}
	boxes {|recursive= true|
		if(recursive, {
			^this.prBoxes(dict['patcher']);
		});
		^dict['patcher']['boxes'];
	}
	lines {|recursive= true|
		if(recursive, {
			^this.prLines(dict['patcher']);
		});
		^dict['patcher']['lines'];
	}
	boxesClass {|maxclass, recursive= true|
		^this.boxes(recursive).select{|x| x['box']['maxclass'].asSymbol==maxclass};
	}
	findBox {|id|									//return box with matching id string
		^this.boxes(true).detect{|x|
			x['box']['id']==id;
		};
	}
	nextId {										//return string with next free object id
		var maxId= 0;
		this.boxes(true).do{|x|
			var id= x['box']['id'].drop(4).asInteger;
			if(id>maxId, {
				maxId= id;
			});
		};
		^"obj-"++(maxId+1);
	}
	makeButton {|rect, extra|
		var id= this.nextId;
		var box= this.prMake("button", id, (rect ? Rect(500.rand, 500.rand, 20, 20)), 1, 1);
		box['box'].put('outlettype', ["bang"]);
		if(extra.notNil, {extra.keysValuesDo{|x, y| box['box'].put(x, y)}});
		dict['patcher']['boxes']= dict['patcher']['boxes'].add(box);
		^id
	}
	makeComment {|rect, text, extra|
		var id= this.nextId;
		var box= this.prMake("comment", id, (rect ? Rect(500.rand, 500.rand, 150, 20)), 1, 0);
		box['box'].put('fontname', "Arial");
		box['box'].put('fontsize', 12);
		box['box'].put('text', text.asString);
		if(extra.notNil, {extra.keysValuesDo{|x, y| box['box'].put(x, y)}});
		dict['patcher']['boxes']= dict['patcher']['boxes'].add(box);
		^id
	}
	makeFlonum {|rect, extra|
		var id= this.nextId;
		var box= this.prMake("flonum", id, (rect ? Rect(500.rand, 500.rand, 50, 20)), 1, 2);
		box['box'].put('outlettype', ["float", "bang"]);
		box['box'].put('fontname', "Arial");
		box['box'].put('fontsize', 12);
		if(extra.notNil, {extra.keysValuesDo{|x, y| box['box'].put(x, y)}});
		dict['patcher']['boxes']= dict['patcher']['boxes'].add(box);
		^id;
	}
	makeLed {|rect, extra|
		var id= this.nextId;
		var box= this.prMake("led", id, (rect ? Rect(500.rand, 500.rand, 20, 20)), 1, 1);
		box['box'].put('outlettype', ["int"]);
		if(extra.notNil, {extra.keysValuesDo{|x, y| box['box'].put(x, y)}});
		dict['patcher']['boxes']= dict['patcher']['boxes'].add(box);
		^id;
	}
	makeMessage {|rect, text, extra|
		var id= this.nextId;
		var box= this.prMake("message", id, (rect ? Rect(500.rand, 500.rand, 60, 18)), 2, 1);
		box['box'].put('outlettype', [""]);
		box['box'].put('fontname', "Arial");
		box['box'].put('fontsize', 12);
		box['box'].put('text', text.asString);
		if(extra.notNil, {extra.keysValuesDo{|x, y| box['box'].put(x, y)}});
		dict['patcher']['boxes']= dict['patcher']['boxes'].add(box);
		^id;
	}
	makeNumber {|rect, extra|
		var id= this.nextId;
		var box= this.prMake("number", id, (rect ? Rect(500.rand, 500.rand, 50, 20)), 1, 2);
		box['box'].put('outlettype', ["int", "bang"]);
		box['box'].put('fontname', "Arial");
		box['box'].put('fontsize', 12);
		if(extra.notNil, {extra.keysValuesDo{|x, y| box['box'].put(x, y)}});
		dict['patcher']['boxes']= dict['patcher']['boxes'].add(box);
		^id;
	}
	makeObject {|rect, text, numInlets= 1, numOutlets= 1, extra|
		var id= this.nextId;
		var box= this.prMake("newobj", id, (rect ? Rect(500.rand, 500.rand, 60, 20)), numInlets, numOutlets);
		box['box'].put('outlettype', [""]);
		box['box'].put('fontname', "Arial");
		box['box'].put('fontsize', 12);
		box['box'].put('text', text.asString);
		if(extra.notNil, {extra.keysValuesDo{|x, y| box['box'].put(x, y)}});
		dict['patcher']['boxes']= dict['patcher']['boxes'].add(box);
		^id
	}
	makePatcher {|rect, maxPat, extra|
		var id, obj;
		rect= (rect ? Rect(500.rand, 500.rand, 60, 20));
		id= this.makeObject(rect, "p"+maxPat.name, 1, 1, extra);
		obj= this.findBox(id);
		obj['box'].put('patcher', maxPat.dict['patcher']);
		^id;
	}
	makePatchLine {|srcId, dstId, srcOutlet= 0, dstInlet= 0, extra|
		var line= (
			'patchline' : (
				'source' : [srcId, srcOutlet],
				'destination' : [dstId, dstInlet],
				'midpoints' : []
			)
		);
		if(extra.notNil, {extra.keysValuesDo{|x, y| line['patchline'].put(x, y)}});
		dict['patcher']['lines']= dict['patcher']['lines'].add(line);
	}
	makeToggle {|rect, extra|
		var id= this.nextId;
		var box= this.prMake("toggle", id, (rect ? Rect(500.rand, 500.rand, 20, 20)), 1, 1);
		box['box'].put('outlettype', ["int"]);
		if(extra.notNil, {extra.keysValuesDo{|x, y| box['box'].put(x, y)}});
		dict['patcher']['boxes']= dict['patcher']['boxes'].add(box);
		^id;
	}
	
	
	
	
	//------------------------------------------------
	//--private - max5
	
	prMake {|maxclass, id, rect, numinlets, numoutlets|
		^(
			'box' : (
				'maxclass' : maxclass,
				'id' : id,
				'patching_rect' : rect.asCollection,
				'numinlets' : numinlets,
				'numoutlets' : numoutlets
			)	
		);
	}
	prBoxes {|d|									//recursively collect boxes in an array
		var boxes= [];
		d['boxes'].do{|x|
			boxes= boxes.add(x);
			if(x['box']['patcher'].notNil, {
				boxes= boxes++this.prBoxes(x['box']['patcher']);
			});
		};
		^boxes;
	}
	prLines {|d|									//recursively collect lines in an array
		var lines= d['lines'];
		d['boxes'].do{|x|
			if(x['box']['patcher'].notNil, {
				lines= lines++this.prLines(x['box']['patcher']);
			});
		};
		^lines;
	}
	prRead {|path|
		var f= File(path, "r");
		var char, line= "", str= "";
		var processLine= {
			var keyStart, keyEnd;
			if(line.includes(58.asAscii), {
				keyStart= line.indexOf(34.asAscii);
				keyEnd= line.find(34.asAscii.asString, false, keyStart+1);
				line.put(keyStart, 39.asAscii).put(keyEnd, 39.asAscii);
				if(line[line.size-2]==123.asAscii, {
					line.put(line.size-2, 40.asAscii);
				});
			}, {
				line= line.tr(123.asAscii, 40.asAscii).tr(125.asAscii, 41.asAscii);
			});
			str= str++line
				.replace("., ", ", ")
				.replace(". "++93.asAscii, " "++93.asAscii)
				.replace(".,"++Char.nl, ","++Char.nl)
				.replace("."++Char.nl, ""++Char.nl);
			if(line.size>8192 and:{line.find("'data'").notNil}, {
				(this.class.name++": found line with >8192 characters. patch might not work. try to remove embedded fpic").warn;
			});
			line= "";
		};
		name= path.basename;
		while({char= f.getChar; char.notNil}, {
			if(char==13.asAscii, {char= Char.nl});
			line= line++char;
			if(char==Char.nl, {processLine.value});
		});
		if(line.size>0, {processLine.value});			//if no newline char at the end
		f.close;
		dict= str.interpret;
	}
	prAsMax5 {|str, d, depth|						//recursively build json maxpat from dict
		var test;
		var ind= Char.tab.dup(depth+1).join;
		d.keysValuesDo{|key, val, index|
			str= str++ind++key.asString.quote+58.asAscii;
			case
				{#['patcher', 'box', 'patchline', 'saved_object_attributes', 'restore'].includes(key)} {
					if(d[key].isArray, {
						str= str++91.asAscii;
						d[key].do{|x, jndex|
							if(x.isString, {
								str= str+x.quote;
							}, {
								str= str+x;
							});
							if(jndex<(d[key].size-1), {str= str++Char.comma});
						};
						str= str+93.asAscii++Char.nl;
						if(index<(d.size-1), {str= str++Char.comma});
					}, {
						str= str+Char.tab+123.asAscii++Char.nl;
						str= this.prAsMax5(str, d[key], depth+1);
						if(index<(d.size-1), {str= str++Char.comma});
					});
				}
				{#['boxes', 'lines', 'preset_data'].includes(key)} {
					str= str+91.asAscii+Char.tab;
					val.do{|x, jndex|
						str= str++123.asAscii++Char.nl;
						str= this.prAsMax5(str, x, depth+1);
						if(jndex<(val.size-1), {str= str++Char.comma});
						str= str+ind++Char.tab;
					};
					str= str++93.asAscii++Char.nl;
					if(index<(d.size-1), {str= str++Char.comma});
				}
				{#['save', 'save2', 'data', 'table_data'].includes(key)} {
					str= str+91.asAscii;
					val.do{|x, jndex|
						if(x.isString, {
							str= str+x.quote;
						}, {
							str= str+x;
						});
						if(jndex<(val.size-1), {str= str++Char.comma});
					};
					str= str+93.asAscii;
					if(index<(d.size-1), {str= str++Char.comma});
					str= str++Char.nl;
				}
				{
					if(val.isString, {
						str= str+this.prEscapes(val).quote;
					}, {
						if(val.isArray, {
							val= val.collect{|x|
								if(x.isString, {
									x.quote;
								}, {
									x;
								});
							};
						});
						str= str+val;
					});
					if(index<(d.size-1), {str= str++Char.comma});
					str= str++Char.nl;
				};
		};
		^str++ind.drop(1)++125.asAscii++Char.nl;
	}
	prEscapes {|str|
		var esc;
		str= str
			.escapeChar(92.asAscii)
			.escapeChar(34.asAscii)
			.replace(10.asAscii.asString, "\\n")
			.replace(13.asAscii.asString, "\\r")
			.replace(9.asAscii.asString, "\\t");
		if(str.includes(195.asAscii), {				//fix missing unicodes in sc, not complete
			esc= 195.asAscii.asString;
			str= str
				.replace(esc++165.asAscii, 140.asAscii)//
				.replace(esc++133.asAscii, 129.asAscii)//
				.replace(esc++164.asAscii, 138.asAscii)//
				.replace(esc++132.asAscii, 128.asAscii)//
				.replace(esc++182.asAscii, 154.asAscii)//
				.replace(esc++150.asAscii, 133.asAscii)//
				.replace(esc++188.asAscii, 159.asAscii)//
				.replace(esc++156.asAscii, 134.asAscii)//
				.replace(esc++169.asAscii, 142.asAscii)//
				.replace(esc++137.asAscii, 131.asAscii)//
				
				.replace(esc++168.asAscii, 143.asAscii)//
				.replace(esc++136.asAscii, 233.asAscii)//
				.replace(esc++170.asAscii, 144.asAscii)//
				.replace(esc++138.asAscii, 230.asAscii)//
				.replace(esc++160.asAscii, 136.asAscii)//
				.replace(esc++128.asAscii, 203.asAscii)//
				.replace(esc++174.asAscii, 148.asAscii)//
				.replace(esc++142.asAscii, 235.asAscii)//
				.replace(esc++175.asAscii, 149.asAscii)//
				.replace(esc++143.asAscii, 236.asAscii)//
				.replace(esc++171.asAscii, 145.asAscii)//
				.replace(esc++139.asAscii, 232.asAscii)//
				
				.replace(esc++167.asAscii, 141.asAscii)//
				.replace(esc++135.asAscii, 130.asAscii)//
				.replace(esc++166.asAscii, 190.asAscii)//
				.replace(esc++134.asAscii, 174.asAscii)//
				.replace(esc++163.asAscii, 139.asAscii)//
				.replace(esc++131.asAscii, 204.asAscii)//
				.replace(esc++184.asAscii, 191.asAscii)//
				.replace(esc++152.asAscii, 175.asAscii)//
				.replace(esc++177.asAscii, 150.asAscii)//
				.replace(esc++145.asAscii, 132.asAscii)//
				.replace(esc++159.asAscii, 167.asAscii)//
				;
		});
		if(str.includes(197.asAscii), {
			esc= 197.asAscii.asString;
			str= str
				.replace(esc++147.asAscii, 207.asAscii)//
				;
		});
		if(str.includes(226.asAscii), {
			esc= 226.asAscii.asString;
			str= str
				.replace(esc++128.asAscii++162.asAscii, 165.asAscii)//
				.replace(esc++130.asAscii++172.asAscii, 219.asAscii)//
				;
		});
		^str;
		
	}
	
	
	//------------------------------------------------
	//--private - max4
	
	prAsMax4 {|d|
		var str= "";
		if(verbose, {"verbose: creating max4 header".postln});
		str= str++this.prMax4Header(d);
		if(verbose, {"verbose: creating max4 boxes".postln});
		d['boxes'].do{|x, i| str= str++this.prMax4Boxes(x['box'], i, d)};
		if(verbose, {"verbose: creating max4 lines".postln});
		d['lines'].do{|x, i| str= str++this.prMax4Lines(x['patchline'], i, d)};
		if(verbose, {"verbose: creating max4 footer".postln});
		str= str++this.prMax4Footer(d);
		^str
	}
	prMax4Header {|d|
		var str= "";
		str= str++"#N vpatcher"+d['rect'][0]+d['rect'][1]+(d['rect'][2]+d['rect'][0]*scalePatWidth)+(d['rect'][3]+d['rect'][1]*scalePatHeight)++z;
		if(d['bglocked']==1, {
			str= str++"#P setbglock 1"++z;
		});
		str= str++this.prFont(d['default_fontname'], d['default_fontsize']);
		if(d['bgcolor'].notNil and:{d['bgcolor']!=(1!4)}, {//add a bgcolor object to set background colour
			str= str++"#P hidden newex 3 23 110 196617 bgcolor"++this.prRGB(d['bgcolor'])++z;
		});
		^str;
	}
	prMax4Boxes {|dd, boxIndex, currentDict|
		var str= "";
		var magic, temp, temp2, cnt, class;
		var hid= " ";
		if(dd['hidden']==1, {
			hid= " hidden ";
		});
		class= dd['maxclass'].asSymbol;
		if(verbose, {("verbose: class:"+class).postln; dd.postln});
		switch(class,
			
			'bpatcher', {
				str= str++"#P"++hid++"bpatcher"++this.prRect2(dd['patching_rect']);
				if(dd['offset'].notNil, {
					str= str+dd['offset'][0].round+dd['offset'][1].round;
				}, {
					str= str+0+0;
				});
				if(dd['name'].notNil, {
					if(stripExtension, {
						str= str+PathName(dd['name']).fileNameWithoutExtension;
					}, {
						str= str+dd['name'];
					});
				}, {
					str= str++" ";
				});
				magic= 0;
				if(dd['border'].notNil, {
					magic= magic.setBit(0, true);
				});
				if(dd['embed'].notNil and:{dd['embed']!=0}, {
					magic= magic.setBit(1, true);
				});
				if(dd['lockeddragscroll']!=1, {		//default changed from max4
					magic= magic.setBit(2, true);
				});
				str= str+magic;
				dd['args'].do{|x|
					str= str+this.prEscapes(x.asString);
				};
				str= str++z;
				if(dd['embed'].notNil and:{dd['embed']!=0}, {
					str= str++this.prAsMax4(dd['patcher']);//recursion for subpatches
					str= str++"#BP pop 0"++z;		//unused??? what is this zero doing???
				});
			},
			
			'button', {
				str= str++"#P"++hid++"button"++this.prRect(dd['patching_rect']);
				if(dd['fgcolor'].notNil, {
					str= str+this.prColorIndex(dd['fgcolor'])++z;
				}, {
					str= str+0++z;
				});
			},
			
			'comment', {
				str= str++this.prFont(dd['fontname'], dd['fontsize']);
				if(dd['linecount'].notNil, {
					str= str++"#P window linecount"+dd['linecount']++z;
					lineCount= true;
				});
				str= str++"#P"++hid++"comment"++this.prRect(dd['patching_rect']);
				magic= (3<<16)+dd['fontsize'].round;	//font number (upper 16bits) not correct but that does not seems to matter. 196617 default
				str= str+magic;
				if(dd['text'].notNil, {
					str= str+this.prEscapes2(dd['text']).escapeChar($;);
				});
				str= str.replace("\\n", Char.nl)++z;
				if(dd['textcolor'].isNil and:{dd['frgb'].notNil}, {
					str= str++"#B color"+this.prColorIndex(dd['frgb'])++z;
				});
				if(dd['textcolor'].notNil, {		//or frgb here? seems to be the same
					str= str++"#B color"+this.prColorIndex(dd['textcolor'])++z;
				});
			},
			
			'dial', {
				str= str++"#P"++hid++"user dial"++this.prRect2(dd['patching_rect']);
				str= str+(dd['size'] ? 128)+(dd['mult'] ? 1)+(dd['min'] ? 0);
				str= str+0;						//unused???
				magic= 223;
				if(dd['vtracking'].notNil, {
					if(dd['vtracking']!=0, {"comment: vtracking!=0".warn});//debug
					magic= magic.setBit(6, false);
				});
				if(dd['clip'].notNil, {
					if(dd['clip']!=0, {"comment: clip!=0".warn});//debug
					magic= magic.setBit(7, false);
				});
				str= str+magic;
				str= str+(dd['degrees'] ? 270);
				str= str+"1 1.";					//needle thickness and needle length not used
				str= str++this.prRGB(dd['fgcolor'] ? (170/255!3))++this.prRGB(dd['bgcolor'] ? (221/255!3));
				str= str+120+120+120;				//shadow colour not used
				str= str++this.prRGB(dd['outlinecolor'] ? (225/255!3))++this.prRGB(dd['needlecolor'] ? (0!3));
				str= str+0+0+0;					//frame colour not used
				str= str++z;
			},
			
			'dropfile', {
				str= str++"#P"++hid++"user dropfile"++this.prRect3(dd['patching_rect']);
				str= str+0;						//unused???
				str= str++z;
			},
			
			'ezadc~', {
				str= str++"#P"++hid++"user ezadc~"++this.prRect3(dd['patching_rect'].at([0, 1])++#[44, 33]);
				str= str+0;
				str= str++z;
			},
			
			'ezdac~', {
				str= str++"#P"++hid++"user ezdac~"++this.prRect3(dd['patching_rect'].at([0, 1])++#[44, 33]);
				str= str+0;
				str= str++z;
			},
			
			'filtergraph~', {
				str= str++this.prFont(dd['fontname'], dd['fontsize']);
				str= str++"#P"++hid++"user filtergraph~"++this.prRect2(dd['patching_rect']);
				magic= 21;						//font number not correct but that does not seems to matter
				str= str+magic;
				str= str+(dd['fontsize'] ? 9).round;
				if(dd['domain'].notNil, {
					str= str++(this.prFloatArray(dd['domain'].round));
				}, {
					str= str+"20. 20000.";
				});
				if(dd['range'].notNil, {
					str= str++this.prFloatArray(dd['range']);
				}, {
					str= str+"0.0625 16.";
				});
				str= str+0;						//output coef on load - unused???
				str= str+(dd['logfreq'] ? 1);
				str= str+(dd['logamp'] ? 1);
				str= str+(dd['phasespect'] ? 0);		//strange bug in max4???
				str= str+0;						//unused???
				str= str+(dd['numdisplay'] ? 1);
				str= str+(dd['dbdisplay'] ? 1);
				str= str++z;
				if(dd['fgcolor'].notNil, {
					str= str++"#X frgb"++this.prRGB(dd['fgcolor'])++z;
				}, {
					str= str++"#X frgb 170 170 170"++z;
				});
				if(dd['bgcolor'].notNil, {
					str= str++"#X brgb"++this.prRGB(dd['bgcolor'])++z;
				}, {
					str= str++"#X brgb 209 209 209"++z;
				});
				if(dd['curvecolor'].notNil, {
					str= str++"#X rgb2"++this.prRGB(dd['curvecolor'])++z;
				}, {
					str= str++"#X rgb2 0 0 0"++z;
				});
				if(dd['markercolor'].notNil, {
					str= str++"#X rgb3"++this.prRGB(dd['markercolor'])++z;
				}, {
					str= str++"#X rgb3 130 130 130"++z;
				});
				if(dd['bordercolor'].notNil, {
					str= str++"#X rgb4"++this.prRGB(dd['bordercolor'])++z;
				}, {
					str= str++"#X rgb4 0 0 0"++z;
				});
				if(dd['bwidthcolor'].notNil, {
					str= str++"#X rgb5"++this.prRGB(dd['bwidthcolor'])++z;
				}, {
					str= str++"#X rgb5 76 108 172"++z;
				});
				if(dd['hbwidthcolor'].notNil, {
					str= str++"#X rgb6"++this.prRGB(dd['hbwidthcolor'])++z;
				}, {
					str= str++"#X rgb6 76 108 172"++z;
				});
				if(dd['hcurvecolor'].notNil, {
					str= str++"#X rgb7"++this.prRGB(dd['hcurvecolor'])++z;
				}, {
					str= str++"#X rgb7 255 22 22"++z;
				});
				if(dd['linmarkers'].notNil, {
					str= str++"#X linmarkers"++this.prFloatArray(dd['linmarkers'])++z;
				}, {
					str= str++"#X linmarkers 5512.5 11025. 16537.5"++z;
				});
				if(dd['logmarkers'].notNil, {
					str= str++"#X logmarkers"++this.prFloatArray(dd['logmarkers'])++z;
				}, {
					str= str++"#X logmarkers 10. 100. 1000. 10000."++z;
				});
				str= str++"#X nfilters"+dd['nfilters']++z;
				dd['setfilter'].clump(14).do{|x|
					str= str++"#X setfilter"++this.prFloatArray(x)++z;
				};
				str= str++"#X done"++z;
			},
			
			'flonum', {
				str= str++this.prFont(dd['fontname'], dd['fontsize']);
				str= str++"#P"++hid++"flonum"++this.prRect(dd['patching_rect'])+dd['fontsize'].round;
				magic= 0;
				if(dd['minimum'].notNil, {
					str= str+dd['minimum'];
					magic= magic.setBit(0, true);
				}, {
					str= str+0;
				});
				if(dd['maximum'].notNil, {
					str= str+dd['maximum'];
					magic= magic.setBit(1, true);
				}, {
					str= str+0;
				});
				if(dd['fontface'].notNil, {
					if(dd['fontface']==1 or:{dd['fontface']==3}, {//can not do italic
						magic= magic.setBit(2, true);
					});
				});
				if(dd['triangle'].notNil, {
					magic= magic.setBit(5, true);
				});
				if(dd['mouseup'].notNil, {
					magic= magic.setBit(6, true);
				});
				if(dd['cantchange'].notNil, {
					magic= magic.setBit(7, true);
				});
				if(dd['bgcolor'].notNil, {
					if(dd['bgcolor'][3]<0.5, {		//transparent if low alpha
						magic= magic.setBit(13, true);
					});
				});
				str= str+magic;
				magic= 3;							//font number not correct but that does not seems to matter
				str= str+magic;
				str= str++this.prRGB(dd['textcolor'] ? (0!3))++this.prRGB(dd['bgcolor'] ? (221/255!3))++this.prRGB(dd['htextcolor'] ? (222/255!3))++this.prRGB(dd['hbgcolor'] ? (0!3));
				str= str++z;
			},
			
			'fpic', {
				if(dd['data'].notNil, {
					(this.class.name++": embedded fpic (vpicture) not supported").warn;
				});
				str= str++"#P"++hid++"user fpic"++this.prRect2(dd['patching_rect']);
				if(dd['pic'].notNil and:{dd['pic'].size>0}, {
					str= str+this.prEscapes(dd['pic']).quote;
				}, {
					str= str++" ";
				});
				str= str+(dd['xoffset'] ? 0).round+(dd['yoffset'] ? 0).round;
				str= str+0;
				str= str+"0. 0 0";					//time offset and internal storage unused???
				if(dd['autofit'].notNil, {
					if(dd['autofit']==1, {
						str= str+3;
					}, {
						str= str+0;
					});
				});
				str= str++z;
			},
			
			'function', {
				str= str++"#P"++hid++"user function"++this.prRect3(dd['patching_rect']);
				str= str+(dd['clickadd'] ? 1)+(dd['clickmove'] ? 1)+(dd['legend'] ? 0)+(dd['outputmode'] ? 0)++z;
				if(dd['pointcolor'].notNil, {		//points
					str= str++"#X frgb"++this.prRGB(dd['pointcolor'])++z;
				}, {
					str= str++"#X frgb 82 82 82"++z;
				});
				if(dd['bgcolor'].notNil, {			//background
					str= str++"#X brgb"++this.prRGB(dd['bgcolor'])++z;
				}, {
					str= str++"#X brgb 204 204 204"++z;
				});
				if(dd['linecolor'].notNil, {		//line segments
					str= str++"#X rgb2"++this.prRGB(dd['linecolor'])++z;
				}, {
					str= str++"#X rgb2 85 85 85"++z;
				});
				if(dd['sustaincolor'].notNil, {		//sustain points
					str= str++"#X rgb3"++this.prRGB(dd['sustaincolor'])++z;
				}, {
					str= str++"#X rgb3 255 255 255"++z;
				});
				if(dd['textcolor'].notNil, {		//legend text
					str= str++"#X rgb4"++this.prRGB(dd['textcolor'])++z;
				}, {
					str= str++"#X rgb4 0 0 0"++z;
				});
				str= str++"#X rgb5 0 0 0"++z;
				if(dd['addpoints'].notNil, {
					dd['addpoints'].clump(3).do{|x|
						str= str++"#X add".scatList(x)++z;//???add dot???!!!FLOAT
					};
				});
				if(dd['domain'].notNil, {
					str= str++"#X domain"+dd['domain']++z;
				}, {
					str= str++"#X domain 1000."++z;
				});
				if(dd['range'].notNil, {
					str= str++"#X range"++this.prFloatArray(dd['range'])++z;
				}, {
					str= str++"#X range"+"0. 1."++z;
				});
				str= str++"#X done"++z;
			},
			
			'gain~', {
				str= str++"#P"++hid++"user gain~"++this.prRect2(dd['patching_rect']);
				str= str+(dd['size'] ? 158);
				if(dd['bgcolor'].notNil, {
					str= str+this.prColorIndex(dd['bgcolor']);
				}, {
					str= str+0;
				});
				str= str+(dd['inc'] ? 1.071519)+(dd['scale'] ? 7.94321)+this.prFloat(dd['interp'] ? 10)++z;
			},
			
			'gswitch', {
				str= str++"#P"++hid++"user gswitch"++this.prRect2(dd['patching_rect']);
				str= str+(dd['int'] ? 0);
				str= str+0++z;					//unused???
			},
			
			'gswitch2', {
				str= str++"#P"++hid++"user gswitch2"++this.prRect2(dd['patching_rect']);
				str= str+(dd['int'] ? 0);
				str= str+0++z;					//unused???
			},
			
			'hint', {
				str= str++"#P"++hid++"user hint"++this.prRect3(dd['patching_rect']);
				str= str+dd['hint'].quote+(dd['delay'] ? 1000)+100;
				str= str+"0 65535 65535 52428 9 3 0";	//unused???
				str= str+(dd['enabled'] ? 1);
				str= str+0+0+0++z;					//unused???
			},
			
			'imovie', {
				str= str++"#P"++hid++"user imovie"++this.prRect2(dd['patching_rect']);
				str= str+1++" "++z;				//border and name unused???
				if(dd['name'].size>0, {
					([dd['name']]++dd['films']).reverse.do{|x|
						str= str++"#M rd"+x++z;
					};
				});
			},
			
			'incdec', {
				str= str++"#P"++hid++"user incdec"++this.prRect2(dd['patching_rect'])+"0 0"++z;
			},
			
			'inlet', {
				if(dd['comment'].notNil, {
					if(dd['comment'].size>0, {
						str= str++"#N comlet"+this.prEscapes2(dd['comment'])++z;
					});
				});
				str= str++"#P"++hid++"inlet"++this.prRect(dd['patching_rect']);
				str= str+0;						//unused???
				str= str++z;
			},
			
			'jit.cellblock', {
				str= str++"#P"++hid++"user jit.cellblock"++this.prRect3(dd['patching_rect']);
				magic= 3;							//font number not correct but that does not seems to matter
				str= str+magic;
				str= str+dd['fontsize'].round;
				str= str+(dd['cols'] ? 10)+(dd['rows'] ? 10)+(dd['colwidth'] ? 45)+(dd['rowheight'] ? 17)+(dd['readonly'] ? 0)+(dd['border'] ? 1)+(dd['grid'] ? 1)+(dd['savemode'] ? 0);
				if(dd['rowhead'].isNil, {
					str= str+0;
				}, {
					str= str+1;
				});
				if(dd['colhead'].isNil, {
					str= str+0;
				}, {
					str= str+1;
				});
				str= str+(dd['vscroll'] ? 1)+(dd['hscroll'] ? 1);
				if(dd['selmode'].notNil, {
					str= str+dd['selmode'];
				}, {
					str= str+1;
				});
				str= str++this.prRGB(dd['fgcolor'] ? (0!3))++this.prRGB(dd['bgcolor'] ? (255/255!3))++this.prRGB(dd['gridlinecolor'] ? (0!3))++this.prRGB(dd['bordercolor'] ? (0!3))++this.prRGB(dd['textcolor'] ? (191/255!3))++this.prRGB(dd['headercolor'] ? (0!3))++this.prRGB(dd['hcellcolor'] ? (#[215, 215, 240]/255));
				str= str+(dd['hsync'] ? 1)+(dd['vsync'] ? 1)+(dd['selsync'] ? 1)+(dd['outmode'] ? 0)+(dd['precision'] ? 4);
				str= str+0+0+(dd['just'] ? 0)++z;	//these zeros unused???
				dd['cellmap'].do{|x|
					str= str++"#X set".scatList(x.collect{|y| if(y.isString, {this.prEscapes(y)}, {y})})++z;
				};
			},
			
			'jit.fpsgui', {
				str= str++this.prFont(dd['fontname'], dd['fontsize']);
				str= str++"#P"++hid++"user jit.fpsgui"++this.prRect(dd['patching_rect']);
				magic= (3<<16)+dd['fontsize'].round;	//font number (upper 16bits) not correct but that does not seems to matter. 196617 default
				str= str+magic;
				str= str+(dd['mode'] ? 0)++z;
			},
			
			'jit.pwindow', {
				str= str++"#P"++hid++"user jit.pwindow"++this.prRect2(dd['patching_rect']);
				str= str+(dd['border'] ? 0).round+(dd['onscreen'] ? 1)+(dd['idlemouse'] ? 0)+(dd['interp'] ? 0)+(dd['doublebuffer'] ? 1)+(dd['depthbuffer'] ? 1);
				str= str++z;
				if(dd['name'].notNil, {
					str= str++"#X name"+dd['name']++z;
				});
			},
			
			'jsui', {
				str= str++"#P"++hid++"user jsui"++this.prRect2(dd['patching_rect']);
				magic= 1;
				if(dd['border'].notNil, {
					if(dd['border']==0, {
						magic= magic.setBit(0, false);
					});
				});
				if(dd['nofsaa'].notNil, {
					if(dd['nofsaa']==1, {
						magic= magic.setBit(1, true);
					});
				});
				str= str+magic+0+0;				//these zeros unused???
				str= str+(dd['filename'] ? 'none');
				dd['jsarguments'].do{|x|
					str= str+this.prEscapes(x.asString);
				};
				str= str++z;
			},
			
			'kslider', {
				str= str++"#P"++hid++"user kslider"++this.prRect(dd['patching_rect']);
				if(dd['patching_rect'][2]<266, {		//check width and select max4 small or large
					str= str+1;					//small
				}, {
					str= str+0;					//large
				});
				str= str+(dd['offset'] ? 36).round(12)+(dd['range'] ? 48).round(12);
				if(dd['patching_rect'][2]<266, {
					str= str+19+7;				//what are these? unused???
				}, {
					str= str+31+12;				//what are these? unused???
				});
				str= str+(dd['mode'] ? 0);
				str= str++this.prRGB(dd['hkeycolor'] ? (128/255!3))++this.prRGB(dd['hkeycolor'] ? (128/255!3))++this.prRGB(dd['whitekeycolor'] ? (1!3))++this.prRGB(dd['blackkeycolor'] ? (0!3))++this.prRGB(dd['bordercolor'] ? (0!3))++z;
			},
			
			'lcd', {
				str= str++"#P"++hid++"user lcd"++this.prRect2(dd['patching_rect']);
				str= str+(dd['local'] ? 1)+(dd['border'] ? 1)+(dd['idle'] ? 0);
				str= str+0;						//max4 onscreen mode unused???
				str= str+(dd['enablesprites'] ? 0)++z;
			},
			
			'led', {
				str= str++"#P"++hid++"user led"++this.prRect2(dd['patching_rect']);
				if(dd['oncolor'].notNil, {
					str= str+this.prColorIndexLed(dd['oncolor']);
				}, {
					str= str+0;					//default is red
				});
				str= str+(dd['blinktime'] ? 150)++z;
			},
			
			'levelmeter~', {
				str= str++"#P"++hid++"user levelmeter~"++this.prRect2(dd['patching_rect']);
				str= str+(dd['interval'] ? 20)+(dd['attack'] ? 0)+(dd['release'] ? 0)+(dd['offset'] ? 0);
				if(dd['range'].notNil, {
					str= str.scatList(dd['range']);
				}, {
					str= str+"-48 12";
				});
				str= str+0+0+0+0+0++z;				//unused???
				str= str++"#X brgb"+this.prRGB(dd['bgcolor'] ? (240/255!3))++z;
				str= str++"#X frgb"+this.prRGB(dd['fgcolor'] ? (200/255!3))++z;
				str= str++"#X rgb2"+this.prRGB(dd['needlecolor'] ? (0!3))++z;
				str= str++"#X rgb3"+this.prRGB(dd['markercolor'] ? (0!3))++z;
				str= str++"#X rgb4"+this.prRGB(dd['bordercolor'] ? (0!3))++z;
				str= str++"#X rgb5"+this.prRGB(dd['overloadcolor'] ? (#[255, 0, 0]/255))++z;
				str= str++"#X rgb6"+this.prRGB(dd['hotcolor'] ? (#[220, 100, 100]/255))++z;
				str= str++"#X rgb7"+this.prRGB(dd['warmcolor'] ? (#[220, 150, 80]/255))++z;
				str= str++"#X rgb8"+this.prRGB(dd['tepidcolor'] ? (#[220, 200, 100]/255))++z;
				str= str++"#X rgb9"+this.prRGB(dd['coolcolor'] ? (#[180, 210, 100]/255))++z;
				str= str++"#X markers".scatList(dd['markers'] ? #[12, 6, 0, -6, -12, -24, -36, -48])+0++z;
				str= str++"#X markersused"+(dd['markersused'] ? 8)++z;
				str= str++"#X done"++z;
			},
			
			'matrixctrl', {
				str= str++"#P"++hid++"user matrixctrl"++this.prRect2(dd['patching_rect']);
				str= str+(dd['cellpict'] ? 'MatrixDefaultCell.pct')+(dd['bkgndpict'] ? 'MatrixDefaultBkgnd.pct');
				str= str+(dd['patching_rect'][2]*scaleBoxWidth).round+(dd['patching_rect'][3]).round;
				str= str+16+16;					//picture size - how can i know this without reading file from disk?
				str= str+(dd['columns'] ? 8)+(dd['rows'] ? 4);
				str= str+((dd['horizontalspacing'] ? 0)+16)+((dd['verticalspacing'] ? 0)+16);
				str= str+(dd['horizontalmargin'] ? 1)+(dd['verticalmargin'] ? 1);
				magic= 48;
				if(dd['one/row'].notNil, {
					if(dd['one/row']==1, {
						magic= magic.setBit(0, true);
					});
				});
				if(dd['one/column'].notNil, {
					if(dd['one/column']==1, {
						magic= magic.setBit(1, true);
					});
				});
				if(dd['one/matrix'].notNil, {
					if(dd['one/matrix']==1, {
						magic= magic.setBit(2, true);
					});
				});
				if(dd['inactiveimage'].notNil, {
					if(dd['inactiveimage']==0, {
						magic= magic.setBit(4, false);
					});
				});
				if(dd['clickedimage'].notNil, {
					if(dd['clickedimage']==0, {
						magic= magic.setBit(5, false);
					});
				});
				if(dd['imagemask'].notNil, {
					if(dd['imagemask']==1, {
						magic= magic.setBit(6, true);
					});
				});
				if(dd['invisiblebkgnd'].notNil, {
					if(dd['invisiblebkgnd']==1, {
						magic= magic.setBit(8, true);
					});
				});
				str= str+magic;
				str= str+(dd['range'] ? 2);
				str= str+5120;					//unused???
				str= str+0;						//active not saved in max4 so unused
				str= str++z;
			},
			
			'message', {
				str= str++this.prFont(dd['fontname'], dd['fontsize']);
				if(dd['linecount'].notNil, {
					str= str++"#P window linecount"+dd['linecount']++z;
					lineCount= true;
				});
				str= str++"#P"++hid++"message"++this.prRect(dd['patching_rect']);
				magic= (3<<16)+dd['fontsize'].round;	//font number (upper 16bits) not correct but that does not seems to matter. 196617 default
				str= str+magic;
				if(dd['text'].notNil, {
					str= str+this.prEscapes2(dd['text']);
				});
				str= str++z;
				if(dd['bgcolor'].notNil, {
					str= str++"#B color"+this.prColorIndex(dd['bgcolor'])++z;
				});
			},
			
			'meter~', {
				str= str++"#P"++hid++"user meter~"++this.prRect3(dd['patching_rect']);
				str= str+(dd['interval'] ? 50);
				str= str+this.prRGB(dd['coldcolor'] ? (#[0, 168, 0]/255));
				str= str+this.prRGB(dd['bgcolor'] ? (103/255!3));
				str= str+this.prRGB(dd['hotcolor'] ? (#[255, 153, 0]/255));
				str= str+this.prRGB(dd['overloadcolor'] ? (#[255, 0, 0]/255));
				str= str+this.prRGB(dd['warmcolor'] ? (#[217, 217, 0]/255));
				str= str+this.prRGB(dd['tepidcolor'] ? (#[153, 186, 0]/255));
				str= str+(dd['numleds'] ? 12)+(dd['dbperled'] ? 3)+(dd['nhotleds'] ? 3)+(dd['nwarmleds'] ? 3)+(dd['ntepidleds'] ? 3)++z;
			},
			
			'multislider', {
				str= str++"#P"++hid++"user multiSlider"++this.prRect2(dd['patching_rect']);
				if(dd['setminmax'].notNil, {
					str= str++this.prFloatArray(dd['setminmax']);
				}, {
					str= str+"-1. 1.";
				});
				str= str+(dd['size'] ? 1);
				magic= 2681;						//magic number 1
				if(dd['orientation'].notNil, {		//vertical or horizontal
					if(dd['orientation']!=0, {"multislider: orientation!=0".warn});//debug
					magic= magic.setBit(0, false);
				});
				if(dd['settype'].notNil, {			//float or integer
					if(dd['settype']!=0, {"multislider: settype!=0".warn});//debug
					magic= magic.setBit(4, false);
				});
				if(dd['setstyle'].notNil, {
					switch(dd['setstyle'],
						0, {magic= magic&2r1111000011111111|(2r1010<<8)},
						1, {magic= magic&2r1111000011111111|(2r1011<<8)},
						2, {magic= magic&2r1111000011111111|(2r1100<<8)},
						3, {magic= magic&2r1111000011111111|(2r1101<<8)},
						4, {magic= magic&2r1111000011111111|(2r1110<<8)},
						5, {magic= magic&2r1111000011111111|(2r1111<<8)},
						{("multislider: magic number 1 setstyle"+dd['setstyle']+"not found").warn}
					);
				});
				str= str+magic;
				magic= 15;						//magic number 2
				if(dd['border_left'].notNil, {
					magic= magic.setBit(0, false);
				});
				if(dd['border_right'].notNil, {
					magic= magic.setBit(1, false);
				});
				if(dd['border_top'].notNil, {
					magic= magic.setBit(2, false);
				});
				if(dd['border_bottom'].notNil, {
					magic= magic.setBit(3, false);
				});
				if(dd['drawpeaks'].notNil, {
					if(dd['drawpeaks']!=1, {"multislider: drawpeaks!=1".warn});//debug
					magic= magic.setBit(4, true);
				});
				if(dd['contdata'].notNil, {
					if(dd['contdata']!=1, {"multislider: contdata!=1".warn});//debug
					magic= magic.setBit(5, true);
				});
				str= str+magic;
				str= str+(dd['signed'] ? 0)+(dd['spacing'] ? 0)+(dd['thickness'] ? 2)+(dd['candycane'] ? 0)+(dd['ghostbar'] ? 2)+(dd['compatibility'] ? 0);
				str= str++z;
				if(dd['slidercolor'].notNil, {
					str= str++"#M frgb"++this.prRGB(dd['slidercolor'])++z;
				}, {
					str= str++"#M frgb 0 0 0"++z;
				});
				if(dd['bgcolor'].notNil, {
					str= str++"#M brgb"++this.prRGB(dd['bgcolor'])++z;
				}, {
					str= str++"#M brgb 255 255 255"++z;
				});
				if(dd['peakcolor'].notNil, {
					str= str++"#M rgb2"++this.prRGB(dd['peakcolor'])++z;
				}, {
					str= str++"#M rgb2 127 127 127"++z;
				});
				if(dd['bordercolor'].notNil, {
					str= str++"#M rgb3"++this.prRGB(dd['bordercolor'])++z;
				}, {
					str= str++"#M rgb3 0 0 0"++z;
				});
				if(dd['candicane2'].notNil, {
					str= str++"#M rgb4"++this.prRGB(dd['candicane2'])++z;
				}, {
					str= str++"#M rgb4 37 52 91"++z;
				});
				if(dd['candicane3'].notNil, {
					str= str++"#M rgb5"++this.prRGB(dd['candicane3'])++z;
				}, {
					str= str++"#M rgb5 74 105 182"++z;
				});
				if(dd['candicane4'].notNil, {
					str= str++"#M rgb6"++this.prRGB(dd['candicane4'])++z;
				}, {
					str= str++"#M rgb6 112 158 18"++z;
				});
				if(dd['candicane5'].notNil, {
					str= str++"#M rgb7"++this.prRGB(dd['candicane5'])++z;
				}, {
					str= str++"#M rgb7 149 211 110"++z;
				});
				if(dd['candicane6'].notNil, {
					str= str++"#M rgb8"++this.prRGB(dd['candicane6'])++z;
				}, {
					str= str++"#M rgb8 187 9 201"++z;
				});
				if(dd['candicane7'].notNil, {
					str= str++"#M rgb9"++this.prRGB(dd['candicane7'])++z;
				}, {
					str= str++"#M rgb9 224 62 37"++z;
				});
				if(dd['candicane8'].notNil, {
					str= str++"#M rgb10"++this.prRGB(dd['candicane8'])++z;
				}, {
					str= str++"#M rgb10 7 114 128"++z;
				});
			},
			
			'newobj', {
				str= str++this.prFont(dd['fontname'], dd['fontsize']);
				if(dd['linecount'].notNil, {
					str= str++"#P window linecount"+dd['linecount']++z;
					lineCount= true;
				});
				if(dd['save'].notNil and:{#['coll', 'counter', 'detonate', 'funbuff', 'in', 'in~', 'menubar', 'out', 'out~', 'prob', 'qlist', 'sfinfo~', 'sfplay~', 'table', 'thispatcher', 'vst~'].includes(dd['text'].split($ )[0].asSymbol).not}, {(this.class.name++": save - trouble in newobj with"+dd['text'].split($ )[0]).warn});
				if(dd['text'].notNil and:{#['coll', 'counter', 'detonate', 'funbuff', 'in', 'in~', 'menubar', 'out', 'out~', 'prob', 'qlist', 'sfinfo~', 'sfplay~', 'table', 'thispatcher', 'vst~'].includes(dd['text'].split($ )[0].asSymbol)}, {//not 100% safe here!!! what other objects?
					switch(dd['text'].split($ )[0].asSymbol,
						'coll', {
							if(dd['text'].notNil, {
								str= str++"#N"+dd['text']++z;
							});
							cnt= 0;
							dd['save'].do{|x|
								if(x==";", {
									str= str++z;
									cnt= 0;
								}, {
									if(cnt==0, {
										if(x.isFloat, {
											str= str++this.prFloat(x);
										}, {
											str= str++x;
										});
									}, {
										if(x.isFloat, {
											str= str+this.prFloat(x);
										}, {
											str= str+x;
										});
									});
									cnt= 1;
								});
							};
						},
						'counter', {
							str= str++"#N"+dd['text']++z;
							if(dd['saved_object_attributes'].notNil, {
								//???
								(this.class.name++": newobj - counter saved_object_attributes").warn;//debug
							}, {
								str= str++"#X flags 0 0"++z;
							});
						},
						'in', {
							str= str++"#N"+dd['text']++z;
						},
						'in~', {
							str= str++"#N"+dd['text']++z;
						},
						'out', {
							str= str++"#N"+dd['text']++z;
						},
						'out~', {
							str= str++"#N"+dd['text']++z;
						},
						'qlist', {
							str= str++"#N qlist"++z;
							if(dd['save'].size>2, {
								temp= dd['save'].copyToEnd(3);
								temp= temp.replace([";", ";"], '\\;');
								temp= temp.replace(["#X", "insert"]);
								str= str++"#X append";
								temp.do{|x|
									if(x.isFloat, {
										str= str+this.prFloat(x);
									}, {
										str= str+x;
									});
								};
								str= str++z;
							});
						},
						'table', {
							str= str++"#N vtable";
							str= str+(dd['saved_object_attributes']['size'] ? 128);
							str= str++this.prRect3(dd['editor_rect'] ? #[100, 100, 300, 300]);
							magic= (dd['showeditor'] ? 0);//default changed from max4
							if(dd['saved_object_attributes']['signed']==1, {
								magic= magic.setBit(1, true);
							});
							if(dd['saved_object_attributes']['notename']==1, {
								magic= magic.setBit(3, true);
							});
							if(dd['saved_object_attributes']['embed']==1, {
								magic= magic.setBit(4, true);
							});
							//skipped resizing for tables here - bit 5 and upwards
							str= str+magic;
							str= str+dd['saved_object_attributes']['range'];
							if(dd['saved_object_attributes']['name'].size>0, {
								str= str+dd['saved_object_attributes']['name'];
							});
							str= str++z;
							if(dd['table_data'].notNil, {
								str= str++"#T set".scatList(dd['table_data'])++z;
							});
						},
						{
							cnt= 0;
							dd['save'].do{|x|
								if(x==";", {
									str= str++z;
									cnt= 0;
								}, {
									if(cnt==0, {
										if(x.isFloat, {
											str= str++this.prFloat(x);
										}, {
											str= str++x;
										});
									}, {
										if(x.isFloat, {
											str= str+this.prFloat(x);
										}, {
											str= str+x;
										});
									});
									cnt= 1;
								});
							};
						}
					);
					str= str++"#P"++hid++"newobj"++this.prRect(dd['patching_rect']);
					magic= (3<<16)+dd['fontsize'].round;//font number (upper 16bits) not correct but that does not seems to matter. 196617 default
					str= str+magic+this.prEscapes(dd['text']).escapeChar(Char.comma)++z;
					if(dd['color'].notNil, {
						str= str++"#B color"+this.prColorIndex(dd['color'])++z;
					});
				}, {
					if(dd['patcher'].notNil, {
						str= str++this.prAsMax4(dd['patcher']);//recursion
						str= str++"#P"++hid++"newobj";
					}, {
						str= str++"#P"++hid++"newex";
					});
					str= str++this.prRect(dd['patching_rect']);
					magic= (3<<16)+dd['fontsize'].round;//font number (upper 16bits) not correct but that does not seems to matter. 196617 default
					str= str+magic;
					if(dd['text'].notNil, {
						str= str+this.prEscapes(dd['text']).escapeChar(Char.comma);
					});
					str= str++z;
					if(dd['text'].notNil, {
						temp= dd['text'].split($ )[0].asSymbol;
						case
							{temp=='pattrstorage'} {
								str= str++"#X client_rect"++this.prRect2(dd['saved_object_attributes']['client_rect'])++z;
								str= str++"#X storage_rect"++this.prRect2(dd['saved_object_attributes']['storage_rect'])++z;
							}
							{temp=='pattr'} {
								str= str++"#X prestore";
								str= str+1+0;		//unused???
								dd['restore'].do{|x|
									str= str+this.prFloat(x);
								};
								str= str++z;
							}
							{#['in', 'in~', 'out', 'out~'].includes(temp)} {
								if(dd['saved_object_attributes'].notNil, {
									if(dd['saved_object_attributes']['attr_comment'].asString.notEmpty, {
										str= str++"#X comment";
										dd['saved_object_attributes']['attr_comment'].do{|x|
											if(x.isFloat, {
												str= str+this.prFloat(x);
											}, {
												if(x.isString, {
													str= str+this.prEscapes2(x);
												}, {
													str= str+x;
												});
											});
										};
										str= str++z;
									});
								});
							};
					});
					if(dd['color'].notNil, {
						str= str++"#B color"+this.prColorIndex(dd['color'])++z;
					});
				});
			},
			
			'nslider', {
				str= str++"#P"++hid++"user nslider"++this.prRect3(dd['patching_rect']);
				str= str+0;						//unused???
				str= str+(dd['mode'] ? 0);
				str= str+0+0++z;					//unused???
			},
			
			'number', {
				str= str++this.prFont(dd['fontname'], dd['fontsize']);
				str= str++"#P"++hid++"number"++this.prRect(dd['patching_rect'])+dd['fontsize'].round;
				magic= 0;
				if(dd['minimum'].notNil, {
					str= str+dd['minimum'];
					magic= magic.setBit(0, true);
				}, {
					str= str+0;
				});
				if(dd['maximum'].notNil, {
					str= str+dd['maximum'];
					magic= magic.setBit(1, true);
				}, {
					str= str+0;
				});
				if(dd['fontface'].notNil, {
					if(dd['fontface']==1 or:{dd['fontface']==3}, {//can not do italic
						magic= magic.setBit(2, true);
					});
				});
				if(dd['format'].notNil, {
					switch(dd['format'],
						1, {magic= magic.setBit(4, true)},
						2, {magic= magic.setBit(10, true)},
						3, {magic= magic.setBit(11, true)},
						4, {magic= magic.setBit(8, true)},
						5, {magic= magic.setBit(12, true)}
					)
				});
				if(dd['triangle'].notNil, {
					magic= magic.setBit(5, true);
				});
				if(dd['mouseup'].notNil, {
					magic= magic.setBit(6, true);
				});
				if(dd['cantchange'].notNil, {
					magic= magic.setBit(7, true);
				});
				if(dd['bgcolor'].notNil, {
					if(dd['bgcolor'][3]<0.5, {		//transparent if low alpha
						magic= magic.setBit(13, true);
					});
				});
				str= str+magic;
				magic= 3;							//font number not correct but that does not seems to matter
				str= str+magic;
				str= str++this.prRGB(dd['textcolor'] ? (0!3))++this.prRGB(dd['bgcolor'] ? (221/255!3))++this.prRGB(dd['htextcolor'] ? (222/255!3))++this.prRGB(dd['hbgcolor'] ? (0!3));
				str= str++z;
			},
			
			'number~', {
				str= str++this.prFont(dd['fontname'], dd['fontsize']);
				str= str++"#P"++hid++"user number~"++this.prRect3(dd['patching_rect'])+dd['fontsize'].round;
				magic= 3;							//font number not correct but that does not seems to matter
				str= str+magic;
				switch(dd['mode'],					//perhaps split these modes - but can not be bothered now
					1, {str= str+1+1},
					2, {str= str+2+1},
					{str= str+3+3}
				);
				magic= 0;
				if(dd['minimum'].notNil, {
					str= str+dd['minimum'];
					magic= magic.setBit(0, true);
				}, {
					str= str+0;
				});
				if(dd['maximum'].notNil, {
					str= str+dd['maximum'];
					magic= magic.setBit(1, true);
				}, {
					str= str+0;
				});
				if(dd['fontface'].notNil, {
					if(dd['fontface']==1 or:{dd['fontface']==3}, {//can not do italic
						magic= magic.setBit(2, true);
					});
				});
				if(dd['triangle'].notNil, {
					magic= magic.setBit(5, true);
				});
				if(dd['mouseup'].notNil, {
					magic= magic.setBit(6, true);
				});
				if(dd['cantchange'].notNil, {
					magic= magic.setBit(7, true);
				});
				if(dd['bgcolor'].notNil, {
					if(dd['bgcolor'][3]<0.5, {		//transparent if low alpha
						magic= magic.setBit(13, true);
					});
				});
				str= str+magic;
				str= str+"0.";					//unused???
				str= str+(dd['interval'] ? 250).round;
				str= str+"0.";					//unused???
				str= str++this.prRGB(dd['textcolor'] ? (0!3))++this.prRGB(dd['bgcolor'] ? (221/255!3))++this.prRGB(dd['htextcolor'] ? (222/255!3))++this.prRGB(dd['hbgcolor'] ? (0!3));
				str= str++z;
			},
			
			'outlet', {
				if(dd['comment'].notNil, {
					if(dd['comment'].size>0, {
						str= str++"#N comlet"+this.prEscapes2(dd['comment'])++z;
					});
				});
				str= str++"#P"++hid++"outlet"++this.prRect(dd['patching_rect']);
				str= str+0;						//unused???
				str= str++z;
			},
			
			'panel', {
				str= str++"#P"++hid++"user panel"++this.prRect2(dd['patching_rect'])++z;
				if(dd['mode']==1, {
					if(dd['grad1'].notNil, {
						str= str++"#X brgb"++this.prRGB(dd['grad1'])++z;
					}, {
						if(dd['grad2'].notNil, {
							str= str++"#X brgb"++this.prRGB(dd['grad2'])++z;
						}, {
							str= str++"#X brgb"++this.prRGB((191/255!3))++z;
						});
					});
				}, {
					str= str++"#X brgb"++this.prRGB(dd['bgcolor'] ? (191/255!3))++z;
				});
				str= str++"#X frgb"++this.prRGB(dd['bordercolor'] ? (0!3))++z;
				str= str++"#X border"+(dd['border'] ? 0)++z;//but default was 1 for max4
				str= str++"#X rounded"+(dd['rounded'] ? 8)++z;//but default was 0 for max4
				str= str++"#X shadow"+(dd['shadow'] ? 0)++z;
				str= str++"#X done"++z;
			},
			
			'pictctrl', {
				str= str++"#P"++hid++"user pictctrl"+this.prRect2(dd['patching_rect']);
				if(dd['name'].notNil, {
					if(dd['name'].includes($ ), {
						str= str+dd['name'].quote;
					}, {
						str= str+dd['name'];
					});
				}, {
					str= str+'<none>';
				});
				str= str+0+0+0;
				str= str+(dd['mode'] ? 0);
				magic= 26;
				if(dd['inactiveimage'].notNil, {
					if(dd['inactiveimage']!=1, {"pictctrl: inactiveimage!=1".warn});//debug
					magic= magic.setBit(0, true);
				});
				if(dd['tracking'].notNil, {			//default changed from max4
					if(dd['tracking']!=1, {"pictctrl: tracking!=1".warn});//debug
					magic= magic.setBit(1, true);
				}, {
					magic= magic.setBit(1, false);
				});
				if(dd['trackhorizontal'].notNil, {	//default changed from max4
					if(dd['trackhorizontal']!=1, {"pictctrl: trackhorizontal!=1".warn});//debug
					magic= magic.setBit(3, true);
				}, {
					magic= magic.setBit(3, false);
				});
				if(dd['trackvertical'].notNil, {		//default changed from max4
					if(dd['trackvertical']!=1, {"pictctrl: trackvertical!=1".warn});//debug
					magic= magic.setBit(4, true);
				}, {
					magic= magic.setBit(4, false);
				});
				if(dd['clickedimage'].notNil, {		//default changed from max4
					if(dd['clickedimage']!=1, {"pictctrl: clickedimage!=1".warn});//debug
					magic= magic.setBit(5, true);
				}, {
					magic= magic.setBit(5, false);
				});
				if(dd['imagemask'].notNil, {
					if(dd['imagemask']!=1, {"pictctrl: imagemask!=1".warn});//debug
					magic= magic.setBit(6, true);
				});
				if(dd['trackcircular'].notNil, {
					if(dd['trackcircular']!=1, {"pictctrl: trackcircular!=1".warn});//debug
					magic= magic.setBit(7, true);
				});
				if(dd['clickincrement'].notNil, {
					if(dd['clickincrement']!=1, {"pictctrl: clickincrement!=1".warn});//debug
					magic= magic.setBit(8, true);
				});
				str= str+magic;
				str= str+0+0;						//unused???
				str= str+(1-(dd['active'] ? 1));
				str= str+0+0;						//half width and full height - how can i know this without reading file from disk?
				str= str+(dd['range'] ? 128);
				str= str+(dd['offset'] ? 0);
				str= str+(dd['multiplier'] ? 1);
				str= str+(dd['frames'] ? 10);
				str= str+(dd['ratio'] ? 2);
				str= str+(dd['snap'] ? 0);
				str= str+1;						//unused???
				str= str+(dd['degrees'] ? 270);
				str= str++z;
			},
			
			'pictslider', {
				str= str++"#P"++hid++"user pictslider"++this.prRect2(dd['patching_rect']);
				str= str+(dd['leftmargin'] ? 4)+(dd['topmargin'] ? 4)+(dd['rightmargin'] ? 4)+(dd['bottommargin'] ? 4);
				if(dd['knobpict'].notNil, {
					if(dd['knobpict'].includes($ ), {
						str= str+this.prEscapes(dd['knobpict']).quote;
					}, {
						str= str+this.prEscapes(dd['knobpict']);
					});
				}, {
					str= str+'SliderDefaultKnob.pct';
				});
				str= str+0;
				if(dd['bkgndpict'].notNil, {
					if(dd['bkgndpict'].includes($ ), {
						str= str+this.prEscapes(dd['bkgndpict']).quote;
					}, {
						str= str+this.prEscapes(dd['bkgndpict']);
					});
				}, {
					str= str+'SliderDefaultBkgnd.pct';
				});
				str= str+0;
				magic= 2163;						//magic number 1
				if(dd['clickedimage'].notNil, {
					magic= magic.setBit(5, false);
				});
				if(dd['imagemask'].notNil, {			//default changed from max4
					magic= magic.setBit(6, true);
				});
				if(dd['scaleknob'].notNil, {
					magic= magic.setBit(8, true);
				});
				if(dd['bkgndsize'].notNil, {
					magic= magic.setBit(9, true);
				});
				if(dd['invisiblebkgnd'].notNil, {
					magic= magic.setBit(10, true);
				});
				str= str+magic;
				str= str+(dd['active'] ? 0);
				magic= 8323072;					//magic number 2
				if(dd['leftvalue'].notNil, {
					magic= magic&0xffff0000|dd['leftvalue'];
				});
				if(dd['topvalue'].notNil, {
					magic= magic&0x0000ffff|(dd['topvalue']<<16);
				});
				str= str+magic;
				magic= 127;
				if(dd['bottomvalue'].notNil, {
					magic= magic&0xffff0000|(dd['bottomvalue']<<16);
				});
				if(dd['rightvalue'].notNil, {
					magic= magic&0xffff0000|dd['rightvalue'];
				});
				str= str+magic;
				str= str+this.prFloat(dd['horizontaltracking'] ? 1)+this.prFloat(dd['verticaltracking'] ? 1);
				str= str++z;
			},
			
			'playbar', {
				str= str++"#P"++hid++"user playbar"++this.prRect2(dd['patching_rect']);
				str= str+0++z;					//unused???
			},
			
			'preset', {
				str= str++"#N vpreset";
				if(dd['preset_data'].notNil, {
					str= str+dd['preset_data'].collect{|x| x['number']}.maxItem++z;
				}, {
					str= str+1++z;
				});
				if(dd['embed'].notNil and:{dd['embed']==0}, {
					str= str++"#X embed 0"++z;
				}, {
					dd['preset_data'].do{|dx|
						str= str++"#X append"+dx['number'];
						str= str+2;				//unused???
						temp= dx['data'];
						cnt= 0;
						while({temp.notEmpty}, {
							if(temp[1]=='<invalid>', {
								//(this.class.name++": preset - skipping invalid preset").warn;//debug
							}, {
								temp2= currentDict['boxes'].detect{|x| x['box']['id']==temp[1]};
								if(temp2.isNil, {
									(this.class.name++": preset - box id"+temp[1]+"not found. skipping").warn;//debug
								}, {
									cnt= cnt+1;
									str= str+cnt;
									temp2= temp2['box'];
									str= str+temp2['patching_rect'][1].round+temp2['patching_rect'][0].round;
									temp.copyRange(2, temp[0]-1).do{|x|
										if(x.isFloat, {
											str= str+this.prFloat(x);
										}, {
											if(x.isString, {
												str= str+this.prEscapes(x);
											}, {
												str= str+x;
											});
										});
									};
									str= str+"\\;";
								});
							});
							temp= temp.copyToEnd(temp[0]);
						});
						str= str++z;
					};
				});
				str= str++"#P"++hid++"preset"+this.prRect2(dd['patching_rect'])++z;
			},
			
			'radiogroup', {
				str= str++"#P"++hid++"user radiogroup"++this.prRect2(dd['patching_rect'])++z;
				str= str++"#X size"+dd['size']++z;
				str= str++"#X offset"+(dd['offset'] ? 16)++z;
				str= str++"#X inactive"+(dd['inactive'] ? 0)++z;
				str= str++"#X itemtype"+dd['itemtype']++z;
				str= str++"#X flagmode"+(dd['flagmode'] ? 0)++z;
				if(dd['itemtype']==0, {				//radio buttons
					str= str++"#X set"+dd['value']++z;
				}, {								//check boxes
					str= str++"#X set".scatList(dd['values'])++z;
				});
				str= str++"#X done"++z;
			},
			
			'rslider', {
				str= str++"#P"++hid++"user rslider"++this.prRect2(dd['patching_rect']);
				str= str+(dd['size'] ? 128).round;
				str= str+(dd['mult'] ? 1).round;
				str= str+0;						//unused???
				if(dd['bgcolor'].notNil, {
					str= str+this.prColorIndex(dd['bgcolor']);
				}, {
					str= str+0;
				});
				str= str+(dd['listmode'] ? 0)++z;
			},
			
			'scope~', {
				str= str++"#P"++hid++"user scope~"++this.prRect3(dd['patching_rect']);
				str= str+(dd['calccount'] ? 256);
				str= str+3;						//unused???
				str= str+(dd['bufsize'] ? 128);
				if(dd['range'].notNil, {
					str= str++this.prFloatArray(dd['range']);
				}, {
					str= str+"-1. 1.";
				});
				str= str+0;						//unused???
				str= str+this.prFloat(dd['delay'] ? 3);
				str= str+(dd['trigger'] ? 0)+this.prFloat(dd['triglevel'] ? 0);
				str= str++this.prRGB(dd['fgcolor'] ? (#[102, 255, 51]/255))++this.prRGB(dd['bgcolor'] ? (135/255!3));
				str= str+(dd['drawstyle'] ? 0)++z;
			},
			
			'slider', {
				if(dd['floatoutput'].notNil, {
					(this.class.name++": slider - floatoutput not supported by max4. replace it with a multislider").warn;
				});
				if(dd['orientation'].isNil, {
					if(dd['patching_rect'][2]<dd['patching_rect'][3], {
						str= str++"#P"++hid++"user uslider"++this.prRect2(dd['patching_rect']);
					}, {
						str= str++"#P"++hid++"user hslider"++this.prRect2(dd['patching_rect'].at(#[0, 1, 3, 2]));
					});
				}, {
					if(dd['orientation']==2, {
						str= str++"#P"++hid++"user uslider"++this.prRect2(dd['patching_rect']);
					}, {
						if(dd['orientation']!=1, {("slider: orientation"+dd['orientation']).warn});
						str= str++"#P"++hid++"user hslider"++this.prRect2(dd['patching_rect'].at(#[0, 1, 3, 2]));
					});
				});
				str= str+(dd['size'] ? 128).round;
				str= str+(dd['mult'] ? 1).round;
				str= str+(dd['min'] ? 0).round;
				if(dd['bgcolor'].notNil, {
					str= str+this.prColorIndex(dd['bgcolor'])++z;
				}, {
					str= str+0++z;
				});
			},
			
			'spectroscope~', {
				str= str++"#P"++hid++"user spectroscope~"++this.prRect2(dd['patching_rect']);
				str= str+(dd['interval'] ? 20);
				str= str+(dd['orientation'] ? 0);
				str= str+(dd['sono'] ? 0);
				str= str+(dd['scroll'] ? 0);
				str= str+(dd['monochrome'] ? 1);
				str= str+(dd['border'] ? 1);
				str= str+(1-(dd['logamp'] ? 0));
				str= str+(dd['logfreq'] ? 0);
				str= str+(dd['phasespect'] ? 0);
				str= str+0+0+0++z;					//unused???
				str= str++"#X frgb"++this.prRGB(dd['fgcolor'] ? (224/255!3))++z;
				str= str++"#X brgb"++this.prRGB(dd['bgcolor'] ? (1!3))++z;
				str= str++"#X rgb2"++this.prRGB(dd['curvecolor'] ? (0!3))++z;
				str= str++"#X rgb3"++this.prRGB(dd['peakcolor'] ? (#[243, 204, 204]/255))++z;
				str= str++"#X rgb4"++this.prRGB(dd['peakcurvecolor'] ? (#[255, 0, 0]/255))++z;
				str= str++"#X rgb5"++this.prRGB(dd['markercolor'] ? (184/255!3))++z;
				str= str++"#X rgb6"++this.prRGB(dd['bordercolor'] ? (0!3))++z;
				str= str++"#X rgb7"++this.prRGB(dd['sonomonofgcolor'] ? (0!3))++z;
				str= str++"#X rgb8"++this.prRGB(dd['sonomonobgcolor'] ? (1!3))++z;
				str= str++"#X rgb9"++this.prRGB(dd['sonohicolor'] ? (#[255, 0, 0]/255))++z;
				str= str++"#X rgb10"++this.prRGB(dd['sonomedhicolor'] ? (#[255, 191, 0]/255))++z;
				str= str++"#X rgb11"++this.prRGB(dd['sonomedcolor'] ? (#[0, 191, 127]/255))++z;
				str= str++"#X rgb12"++this.prRGB(dd['sonomedlocolor'] ? (#[127, 0, 127]/255))++z;
				str= str++"#X rgb13"++this.prRGB(dd['sonolocolor'] ? (0!3))++z;
				str= str++"#X range"++this.prFloatArray(dd['range'] ? #[0, 1.2])++z;
				str= str++"#X domain"++this.prFloatArray(dd['domain'] ? #[0, 22050])++z;
				str= str++"#X done"++z;
			},
			
			'suckah', {
				str= str++"#P"++hid++"user suckah"++this.prRect2(dd['patching_rect'])++z;
			},
			
			'swatch', {
				str= str++"#P"++hid++"user swatch"++this.prRect2(dd['patching_rect'])++z;
			},
			
			'textedit', {
				str= str++this.prFont(dd['fontname'], dd['fontsize']);
				str= str++"#P"++hid++"user textedit"++this.prRect3(dd['patching_rect']);
				magic= 32896;
				if(dd['lines'].notNil, {			//max number of lines (upper 16 bits)
					magic= magic|(dd['lines']<<16);
				});
				if(dd['readonly'].notNil, {
					magic= magic.setBit(0, true);
				});
				if(dd['clickmode'].notNil, {
					if(dd['clickmode']==1, {
						magic= magic.setBit(1, true);
					});
				});
				if(dd['keymode'].notNil, {
					if(dd['keymode']==1, {
						magic= magic.setBit(2, true);
					});
				});
				if(dd['outputmode'].notNil, {
					if(dd['outputmode']==1, {
						magic= magic.setBit(3, true);
					});
				});
				if(dd['autoscroll'].notNil, {
					if(dd['autoscroll']==0, {
						magic= magic.setBit(4, true);
					});
				});
				if(dd['wordwrap'].notNil, {
					if(dd['wordwrap']==0, {
						magic= magic.setBit(5, true);
					});
				});
				if(dd['tabmode'].notNil, {
					if(dd['tabmode']==0, {
						magic= magic.setBit(7, false);
					});
				});
				str= str+magic;
				magic= 3;							//font number not correct but that does not seems to matter
				str= str+magic;
				str= str+dd['fontsize'].round++z;
			},
			
			'toggle', {
				str= str++"#P"++hid++"toggle"++this.prRect(dd['patching_rect']);
				str= str+0;						//unused???
				str= str++z;
			},
			
			'umenu', {
				str= str++this.prFont(dd['fontname'], dd['fontsize']+2);
				if(dd['menumode'].isNil or:{dd['menumode']==3}, {//ubumenu for normal and toggle
					str= str++"#P"++hid++"user ubumenu"++this.prRect(dd['patching_rect']);
					magic= (3<<16)+dd['fontsize'].round;//font number (upper 16bits) not correct but that does not seems to matter. 196617 default
					str= str+magic;
					if(dd['menumode']==3, {
						str= str+1;
					}, {
						str= str+0;
					});
					str= str+(dd['arrow'] ? 1);
					str= str+1;					//autosize unused???
					str= str+(dd['align'] ? 0);
					str= str++z;
					if(dd['bgcolor'].notNil or:{dd['textcolor'].notNil or:{dd['bgcolor2'].notNil or:{dd['framecolor'].notNil} or:{dd['togcolor'].notNil or:{dd['discolor'].notNil}}}}, {
						str= str++"#X setrgb"
						++this.prRGB(dd['textcolor'] ? (31/255!3))
						++this.prRGB(dd['bgcolor'] ? (187/255!3))
						++this.prRGB(dd['bgcolor2'] ? (234/255!3))
						++this.prRGB(dd['framecolor'] ? (4/255!3))
						++this.prRGB(dd['togcolor'] ? (141/255!3))
						++this.prRGB(dd['discolor'] ? (112/255!3));
						str= str++z;
					});
					
					if(dd['items'].asCollection.notEmpty, {
						str= str++"#X add";
						dd['items'].asCollection.do{|x|
							if(x!=",", {
								str= str+this.prEscapes(x.asString);
							}, {
								str= str++z++"#X add";
							});
						};
						str= str++z;
					});
					if(dd['types'].notEmpty, {
						str= str++"#X types";
						dd['types'].do{|x|
							str= str+this.prEscapes(x);
						};
						str= str++z;
					});
					str= str++"#X prefix_set";
					if(dd['prefix_mode'].notNil and:{dd['prefix_mode']>1}, {
						(this.class.name++": umenu - prefix_mode>1 not supported by max4").warn;
					});
					str= str+(dd['prefix_mode'] ? 0);
					str= str+(dd['autopopulate'] ? 0);
					if(dd['prefix'].notNil, {
						if(dd['prefix'].includes($ ), {
							str= str+dd['prefix'].quote;
						}, {
							str= str+dd['prefix'];
						});
					}, {
						str= str+'<none>';
					});
					str= str+(dd['depth'] ? 0);
					str= str++z;
				}, {								//umenu for scrolling and label
					str= str++"#P"++hid++"user umenu"++this.prRect(dd['patching_rect']);
					magic= (3<<16)+(dd['fontsize'].round.asInteger<<2)+3;//font number (upper 16bits) not correct but that does not seems to matter. 196647 default
					str= str+magic;
					if(dd['menumode']==1, {
						str= str+2;
					}, {
						str= str+3;
					});
					str= str+dd['items'].size.max(64);
					str= str+(dd['patching_rect'][1]+dd['patching_rect'][3]-2).round;
					str= str+(dd['labelclick'] ? 0);	//default changed from max4
					str= str++z;
					if(dd['textcolor'].notNil or:{dd['bgcolor'].notNil or:{dd['framecolor'].notNil}}, {
						str= str++"#X setrgb"
						++this.prRGB(dd['textcolor'] ? (221/255!3))
						++this.prRGB(dd['bgcolor'] ? (0!3))
						++this.prRGB(dd['framecolor'] ? (1!3))//all upper/lower are mapped to just framecolor
						++this.prRGB(dd['framecolor'] ? (221/255!3))
						++this.prRGB(dd['framecolor'] ? (170/255!3))
						++this.prRGB(dd['framecolor'] ? (119/255!3))
						++this.prRGB(dd['framecolor'] ? (187/255!3));
						str= str++z;
					});
					if(dd['items'].asCollection.notEmpty, {
						str= str++"#X add";
						dd['items'].asCollection.do{|x|
							if(x!=",", {
								str= str+this.prEscapes(x.asString);
							}, {
								str= str++z++"#X add";
							});
						};
						str= str++z;
					});
				});
			},
			
			'ubutton', {
				str= str++"#P"++hid++"user ubutton"++this.prRect2(dd['patching_rect']);
				str= str+(dd['stay'] ? 0);
				magic= 0;
				if(dd['toggle'].notNil, {
					if(dd['toggle']==1, {
						magic= magic.setBit(0, true);
					});
				});
				if(dd['hilite'].notNil, {
					if(dd['hilite']==0, {
						magic= magic.setBit(1, true);
					});
				});
				if(dd['dragtrack'].notNil, {
					if(dd['dragtrack']==1, {
						magic= magic.setBit(2, true);
					});
				});
				str= str+magic++z;
			},
			
			'waveform~', {
				if(dd['buffername'].notNil, {
					(this.class.name++": waveform~ - embedded buffer name not supported by max4. set this manually with a set message").warn;
				});
				str= str++this.prFont(dd['fontname'], dd['fontsize']);
				str= str++"#P"++hid++"user waveform~"++this.prRect2(dd['patching_rect']);
				magic= 3;							//font number not correct but that does not seems to matter
				str= str+magic;
				str= str+(dd['fontsize'] ? 9).round++z;
				str= str++"#W mode"+switch(dd['setmode'],
					0, 'none',
					1, 'select',					//default in max4
					2, 'loop',
					3, 'move',
					4, 'draw',
					'none'						//not correct but there is no way to know
				)++z;
				if(dd['snapto'].notNil, {
					str= str++"#W snap"+switch(dd['snapto'],
						0, 'none',
						1, 'grid',
						2, 'zero'
					)++z;
				});
				if(dd['voffset'].notNil and:{dd['voffset']!=0}, {
					str= str++"#W offset"+dd['voffset']++z;
				});
				str= str++"#W mouseoutput"+switch(dd['outmode'],
					0, 'none',
					1, 'down',
					2, 'up',
					3, 'downup',
					4, 'continuous',
					'continuous'
				)++z;
				str= str++"#W unit"+switch(dd[''],
					0, 'ms',
					1, 'samples',
					2, 'phase',
					3, 'bpm',
					'ms'
				)++z;
				str= str++"#W grid"+this.prFloat(dd['grid'] ? 1000)++z;
				str= str++"#W ticks"+(dd['ticks'] ? 0)++z;
				str= str++"#W labels"+(dd['labels'] ? 1)++z;
				str= str++"#W vlabels"+(dd['vlabels'] ? 0)++z;
				str= str++"#W vticks"+(dd['vticks'] ? 1)++z;
				str= str++"#W bpm"+this.prFloat(dd['attr_bpm'] ? 120)+this.prFloat(dd['beats'] ? 4)++z;
				str= str++"#W frgb"+this.prRGB(dd['waveformcolor'] ? (#[33, 0, 0]/255))++z;
				str= str++"#W brgb"+this.prRGB(dd['bgcolor'] ? (#[60, 178, 173]/255))++z;
				str= str++"#W rgb2"+this.prRGB(dd['selectioncolor'] ? (#[0, 95, 255]/255))++z;
				str= str++"#W rgb3"+this.prRGB(dd['bordercolor'] ? (#[0, 0, 0]/255))++z;
				str= str++"#W rgb4"+this.prRGB(dd['labeltextcolor'] ? (#[0, 0, 0]/255))++z;
				str= str++"#W rgb5"+this.prRGB(dd['labelbgcolor'] ? (#[190, 137, 255]/255))++z;
				str= str++"#W rgb6"+this.prRGB(dd['tickmarkcolor'] ? (100/255!3))++z;
				if(dd['selectioncolor'].notNil, {
					str= str++"#W rgb7"+this.prRGB(dd['selectioncolor'][3]!3)++z;
				}, {
					str= str++"#W rgb7 100 100 100"++z;
				});
			},
			
			'zplane~', {
				str= str++"#P"++hid++"user zplane~"++this.prRect2(dd['patching_rect'])++z;
				str= str++"#X rgb 224 224 224"++z;
				str= str++"#X pconstrain 0"++z;
				str= str++"#X done"++z;
			},
			
			//--object not found
			{
				(this.class.name++": maxclass '"++class++"' not found! replacing with a red dummy bpatcher").warn;
				str= str++"#P bpatcher"++this.prRect2(dd['patching_rect'])+0+0+"dummy"+2++z;
				str= str++"#N vpatcher 0 0 999 999"++z;
				dd['numoutlets'].do{|x, i|
					str= str++"#P outlet"+(i*30+10)+72+16+0++z;
				};
				dd['numinlets'].do{|x, i|
					str= str++"#P inlet"+(i*30+10)+36+16+0++z;
				};
				str= str++"#P comment 0 0 200 196617 REPLACE"+class++z;
				str= str++"#P user panel 0 0 999 999"++z;
				str= str++"#X brgb 255 0 0"++z;
				str= str++"#X frgb 0 0 0"++z;
				str= str++"#X border 0"++z;
				str= str++"#X rounded 0"++z;
				str= str++"#X shadow 0"++z;
				str= str++"#X done"++z;
				str= str++"#P pop"++z;
				str= str++"#BP pop 0"++z;
			}
		);
		
		if(dd['ignoreclick'].notNil, {
			if(dd['ignoreclick']!=1, {
				//(this.class.name++": ignoreclick!=1").warn;//debug
			}, {
				str= str++"#P noclick"++z;
			});
		});
		if(dd['background'].notNil, {
			if(dd['background']!=1, {
				//(this.class.name++": background!=1").warn;//debug
			}, {
				str= str++"#P background"++z;
			});
		});
		if(dd['varname'].notNil, {
			str= str++"#P objectname"+dd['varname']++z;
		});
		if(lineCount, {
			str= str++"#P window linecount 1"++z;
			lineCount= false;
		});
		dd.put('f0index', boxIndex);
		^str
	}
	prMax4Lines {|dd, lineIndex, currentDict|
		var str= "";
		var srcIndex, srcBox, srcId, dstIndex, dstBox, dstId, points, hid;
		srcIndex= dd['source'];
		srcBox= currentDict['boxes'].detect{|x| x['box']['id']==srcIndex[0]}['box'];
		srcId= currentDict['boxes'].size-srcBox['f0index']-1;
		dstIndex= dd['destination'];
		if(dstIndex[0].size>0, {
			dstBox= currentDict['boxes'].detect{|x| x['box']['id']==dstIndex[0]}['box'];
			dstId= currentDict['boxes'].size-dstBox['f0index']-1;
			points= dd['midpoints'];
			hid= " ";
			if(dd['hidden']==1, {
				hid= " hidden ";
			});
			if(points.size==0, {
				str= str++"#P"++hid++"connect"+srcId+srcIndex[1]+dstId+dstIndex[1]++z;
			}, {
				str= str++"#P"++hid++"fasten"+srcId+srcIndex[1]+dstId+dstIndex[1];
				points.do{|x| str= str+x.round};
				str= str++z;
			});
			if(dd['color'].notNil, {
				str= str++"#P lcolor"+(this.prColorIndex(dd['color'])+1)++z;
			});
		}, {
			(this.class.name++": skipping patchline with no destination").warn;
		});
		^str;
	}
	prMax4Footer {|d|
		^"#P pop"++z;
	}
	
	prFloat {|val|								//returns a string with . at the end
		var str= val.asString;
		if(str.includes($.).not, {
			str= str++$.;
		});
		^str;
	}
	prFloatArray {|arr|							//returns a string from an array of floats
		var str= "";
		arr.do{|x| str= str+this.prFloat(x)};
		^str;
	}
	prRect {|arr|									//return a string with x, y, x1
		^"".scatList((arr*[1, 1, scaleBoxWidth]).keep(3).round);
	}
	prRect2 {|arr|								//return a string with x, y, x1, y1
		^this.prRect(arr.keep(3))+(arr[3]*scaleBoxWidth).round;
	}
	prRect3 {|arr|								//return a string with x, y, w, h
		^this.prRect2([arr[0], arr[1], arr[2]+arr[0], arr[3]+arr[1]]);
	}
	prFont {|name, size|							//tries to match max4 fontname and fontsize
		var str= "";
		if(name.notNil, {
			if(lastFontName!=name or:{lastFontSize!=size}, {
				lastFontName= name;
				lastFontSize= size;
				name= name//.asString
					.replace(" Bold Italic")
					.replace(" Condensed Bold")
					.replace(" Bold")
					.replace(" SemiBold Italic")
					.replace(" Italic")
					.replace(" SemiBold")
					.replace(" Medium")
					.replace(" ExtraBold")
					.replace(" Condensed Black")
					.replace(" Wide")
					.replace(" Thin");				//replacing like this not 100% safe
				if(name.includes($ ), {
					str= "#P window setfont"+name.quote;
				}, {
					str= "#P window setfont "++name;
				});
				str= str+size.round++z;
			});
		});
		^str;
	}
	prRGB {|arr|									//format rgb colours
		^"".scatList((arr.keep(3)*255).round);
	}
	prColorIndex {|rgb|							//return the closest match of the max4 default 16 colours
		var minIndex= 0, minVal= 9999;
		^if(rgb.keep(3)==#[0, 0, 0], {
			0;
		}, {
			(max4colors/255).do{|x, i|
				var diff= (rgb[0]-x[0]).hypot((rgb[1]-x[1]).hypot(rgb[2]-x[2]));
				if(diff<minVal, {
					minVal= diff;
					minIndex= i;
				});
			};
			^minIndex;
		});
	}
	prColorIndexLed {|rgb|							//return the closest match of the max4 default 5 colours
		var minIndex= 0, minVal= 9999;
		(max4colorsLed/255).do{|x, i|
			var diff= (rgb[0]-x[0]).hypot((rgb[1]-x[1]).hypot(rgb[2]-x[2]));
			if(diff<minVal, {
				minVal= diff;
				minIndex= i;
			});
		};
		^minIndex;
	}
	prEscapes2 {|str|								//also does comma, dollar etc
		^this.prEscapes(str)
			.escapeChar($$)
			.replace(";\\r", " \\; ")
			.replace(";\\n#", ";#")
			.replace(",", " \\,")
			.replace(" \\,", " \\, ");
	}
}
